# -*- coding: utf-8 -*-

"""Top-level package for MemCNN."""

__author__ = """Sil van de Leemput"""
__email__ = 'silvandeleemput@gmail.com'
__version__ = '0.3.5'


from memcnn.models.revop import ReversibleBlock

__all__ = [
    'ReversibleBlock'
]
